#!/usr/bin/env python3

import re
import sys
import os
import getopt
import pypint
import subfunctions
import networkx as nx

from time import time

def tic(msg):
    print("# %s" % msg)
    return time()
def toc(t):
    print("# done in %ss" % (time() - t))

def main(argv):

	inputfile = ''
	outputfile = 'dot_files/'
	nb_muts = 2
	permanent = False
	inevitable = True
	fixed_points= False
	excld=""
	only=""
	

	final_state=''

#0	#get opts	
	try:
		opts, args = getopt.getopt(argv,"hi:n:p:e:f:c:x:o:",["ifile=", "nMuts=","isPermanent=", "isExistential=","fixedPoints=","finalStateConditions=","exclude=","only="])
	except getopt.GetoptError:
		print ("""findMutationsPaths.py -i <inputfile> -n <number of mutations> -p <1 if mutations are permanent, 0 otherwise> -e <0 for inevitable paths, 1 for existential ones> -f <0 if no fixed point computing, 1 otherwise> -c <string of conditions> -x <conditions on nodes that can not be muted> -o <conditions on nodes that are the only one that can be muted>
	default values are fixed points are not computed, 2 mutations, temporary ones, and inevitable paths
	the conditions are expressed as "a1=1, a2=1, a3=0, a4=1" """)
		sys.exit(2)
	for opt, arg in opts:
		if opt == '-h':
			print ("""findMutationsPaths.py -i <inputfile> -n <number of mutations> -p <1 if mutations are permanent, 0 otherwise> -e <0 for inevitable paths, 1 for existential ones> -f <0 if no fixed point computing, 1 otherwise> -c <string of conditions> -x <string of nodes that can not be muted (o=0) or are the only one that can be muted (o=1)> -o <0 or 1, see -x, default=0>
	default values are fixed points are not computed, 2 mutations, temporary ones, and inevitable paths
	the conditions are expressed as "a1=1, a2=1, a3=0, a4=1" """)
			sys.exit()
		elif opt in ("-i", "--ifile"):
			inputfile = arg
		elif opt in ("-n", "--nmuts"):
			nb_muts = int(arg)
		elif opt in ("-p", "--ispermanent"):
			permanent= bool(int(arg))
		elif opt in ("-e", "--isexistential"):
			inevitable= not bool(int(arg))
		elif opt in ("-f", "--fixedPoints"):
			fixed_points= bool(int(arg))
		elif opt in ("-c", "--finalStateConditions"):
			final_state= arg
		elif opt in ("-x", "--exclude"):
			excld=re.split('\W+',arg)
		elif opt in ("-o", "--only"):
			only=re.split('\W+',arg)

	splited = re.split('\W+', os.path.basename(inputfile))		#creating a nice outputfile
	for part in splited[:-1]:
		outputfile+=part
	outputfile+="_"+str(nb_muts)
	if permanent:
		outputfile+="pMuts.dot"
	else:
		outputfile+="tMuts.dot"
	
#I	#New Automata with muts
	(muted_file, initial_state) = subfunctions.write_muted_automata(nb_muts, permanent, inputfile, excld, only)
	print("muted automata written in "+muted_file)

	
#II	#State graph computation and simplification
	automata_Network = pypint.load(muted_file).having(**initial_state)
	an = automata_Network

	t = tic("Computing reachable state graph")
	state_graph = automata_Network.reachable_stategraph()
	stg = state_graph
	toc(t)
	print("%d reachable states" % len(state_graph))

	#reducing a bit the graph, the deleted nodes are useless
	t = tic("Parsing graph")
	final_state = subfunctions.parse_label(final_state)
	i_states, final_states = subfunctions.parse_perturbation_stg(an, stg, initial_state, final_state)
	toc(t)

	#subfunctions.dump_graph(stg, "/tmp/stg.dot", i_states, final_states, show_label=False)


#III	#Reachability
		#Validating nodes: make sure each node comming to the final states are 'valid' (meaning it only leads to final states), then check all of their parents, and so on until you have all valid nodes
	t = tic("Graph condensation and red arrows reconstruction")
	condensed_state_graph =subfunctions.graph_wrap(state_graph)		#condensating state graph to avoid issues with loops
	sccg = condensed_state_graph
	toc(t)
	t = tic("Mapping final and initial states")
	final_states = set(final_states)
	final_states_condensed = [node_c for (node_c,data) in \
			condensed_state_graph.nodes(data=True) \
				if data["members"].intersection(final_states)]

	i_states = set(i_states)
	i_states_condensed= [node_c for (node_c,data) in \
			condensed_state_graph.nodes(data=True) \
				if data["members"].intersection(i_states)]
	toc(t)

	#reducing final states to attractors
	if fixed_points:
		tic("Reducing final states to attractors")
		def is_final_scc(n):
		    #if all outgoing arrows are red, then it's a fixed point
		    for (u,v,data) in sccg.out_edges(nbunch=n, data=True):
			    if data.get("color") != "red":
				    return False
		    return True
		final_states_condensed = [n for n in final_states_condensed if is_final_scc(n)]
		toc(t)

	#subfunctions.dump_graph(stg, "/tmp/sccg.dot", i_states_condensed, final_states_condensed)

	t = tic("Reachability computation (inevitable=%s)" % inevitable)
	final_valid, continue_algo = subfunctions.reach(condensed_state_graph, i_states_condensed, final_states_condensed, inevitable)	#inevitable reach function, modifies the graph to show the valid nodes, in purple
	toc(t)


	if not continue_algo:
	    print("NO SOLUTION")
	    return

	t = tic("Final graph reduction")
	reduced_graph = subfunctions.graph_reduction(condensed_state_graph, i_states_condensed, final_states_condensed, final_valid)
	toc(t)

	"""
	for n in final_valid:
		for m in reduced_graph.node[n]["members"]:
			print(state_graph.node[m]["label"])
	print(reduced_graph.edges(data=True))
	for n in reduced_graph.nodes():
		for m in reduced_graph.node[n]["members"]:
			print("%s = %s" % (n,state_graph.node[m]["label"]))
	"""

#IV	#Get perturbations and an example of a perturbation path
	t=tic("Finding reprogramming solutions")
	(s, init_s) = subfunctions.get_mutations(reduced_graph, i_states_condensed, final_valid)
	toc(t)

	print("### mutations from initial state")
	for e in sorted(init_s):
		print("	- %s" % e)

	print("### temporal mutations")
	for i,e in enumerate(s):
		print("	%d. %s" % (i+1,e))

	if len(s)>0:
		input_wanted_path=int(input("Which one of these temporal mutations sequences are you interested in? (enter the number in the list, or 0 if you are not interested in an example)\n"))
		if input_wanted_path>0 and input_wanted_path < len(s)+1:
			wanted_path=s[input_wanted_path-1]
			path_graph = subfunctions.find_a_mutation_path(reduced_graph, i_states_condensed, final_valid, wanted_path)
			subfunctions.dump_graph(path_graph, "./dot_files/example_path.dot", [], [], False)
			print("An example of a path have been written in ./dot_files/example_path.dot")



if __name__ == "__main__":
   main(sys.argv[1:])

# vi:set noet:
